﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Toolkit.Algorithm {
	/// <summary>
	/// 最急降下法
	/// </summary>
	public class SteepestDescent {
		/// <summary>
		/// 具体的な処理の実装
		/// </summary>
		public interface IStrategy {
			/// <summary>
			/// 初期化。
			/// </summary>
			bool Initialize(out double[] parameters, out double alpha);
			/// <summary>
			/// 前処理
			/// </summary>
			/// <param name="sender">呼び出し元</param>
			/// <returns>falseを返したらこのステップは実行しない。</returns>
			bool StartStep(SteepestDescent sender);
			/// <summary>
			/// i番目の変数で偏微分した関数にpを与えたときの値
			/// </summary>
			/// <example>
			/// 実装例
			/// <code>
			///	const double h = 0.01; // 差分幅
			/// double Diff(SteepestDescent sender, int i) {
			///		double f0 = f(sender.Parameters); // これはDoStep()実行前に計算しとくと良い。
			///		sender.Parameters[i] += h;
			///		double f1 = f(sender.Parameters)
			///		sender.Parameters[i] -= h; // 戻しとかないとバグるので注意
			///		return (f1 - f0) / h;
			/// }
			/// double f(double[] parameters) {...} // 最適化したい関数
			/// </code>
			/// </example>
			/// <param name="sender">呼び出し元</param>
			/// <param name="i">序数</param>
			/// <returns>関数値</returns>
			double Diff(SteepestDescent sender, int i);
		}

		IStrategy strategy;
		double[] parameters = null;

		double[] alphaArray = null;
		int[] signs = null;

		/// <summary>
		/// 変数の値
		/// </summary>
		public double[] Parameters {
			get { return parameters; }
		}

		/// <summary>
		/// 値の更新幅。既定値は1.0。(適当)
		/// 負の数にすると最大値に集束する。(正の数なら最小値に集束。「降下法」なので。)
		/// </summary>
		public double Alpha { get; set; }

		/// <summary>
		/// Alphaを値毎に持って、微分値の符号が反転したら更新幅を減らしていくアルゴリズムを使用するかどうか。
		/// 最急降下法を独自にアレンジしてみたもの。
		/// 同じのとかもっといいのとかあるかも知らんけど。
		/// </summary>
		public bool UseSeparateAlpha { get; set; }

		/// <summary>
		/// 初期化
		/// </summary>
		/// <param name="strategy">IStrategy</param>
		public SteepestDescent(IStrategy strategy) {
			this.strategy = strategy;
		}

		/// <summary>
		/// 処理を1ステップ行う
		/// </summary>
		public void DoStep() {
			if (parameters == null) {
				double alpha;
				if (!strategy.Initialize(out parameters, out alpha)) return;
				Alpha = alpha;
			}
			if (UseSeparateAlpha) {
				alphaArray = new double[parameters.Length];
				signs = new int[parameters.Length];
				for (int i = 0; i < parameters.Length; i++) alphaArray[i] = Alpha;
			} else {
				alphaArray = null;
				signs = null;
			}
			double[] diff = new double[parameters.Length];
			// 前処理
			if (!strategy.StartStep(this)) return;
			// 微分値を取得
			for (int i = 0; i < parameters.Length; i++) {
				double d = strategy.Diff(this, i);
				System.Diagnostics.Debug.Assert(!double.IsNaN(d), "strategy.Diff()がdouble.NaN返した。計算ミス？");
				diff[i] = d;
			}
			// パラメータの更新
			if (UseSeparateAlpha) {
				for (int i = 0; i < parameters.Length; i++) {
					if (double.IsNaN(diff[i])) continue;
					// 符号反転時には幅を少し減らす。
					if (diff[i] < 0 && 0 < signs[i]) alphaArray[i] *= 0.9;
					if (signs[i] < 0 && 0 < diff[i]) alphaArray[i] *= 0.9;
					// 符号の記録
					signs[i] =
						diff[i] < 0 ? -1 : 
						0 < diff[i] ? +1 : 0;
					// 値の補正
					parameters[i] -= diff[i] * alphaArray[i];
				}
			} else {
				for (int i = 0; i < parameters.Length; i++) {
					if (double.IsNaN(diff[i])) continue;
					parameters[i] -= diff[i] * Alpha;
				}
			}
		}

		/// <summary>
		/// UseSeparateAlpha時のAlpha値の最小値を返す。
		/// </summary>
		public double GetMinAlphaAbs() {
			if (alphaArray == null) {
				return double.NaN;
			}
			double min = double.MaxValue;
			for (int i = 0; i < alphaArray.Length; i++) {
				if (alphaArray[i] < min) min = Math.Abs(alphaArray[i]);
			}
			return min;
		}

		/// <summary>
		/// パラメータを文字列化。
		/// CVSファイルに入れる用に、, 区切りで前後には何も付けない。
		/// </summary>
		public string ParametersToString() {
			return ParametersToString(",");
		}
		/// <summary>
		/// パラメータを文字列化
		/// </summary>
		/// <param name="separator">セパレータ</param>
		public string ParametersToString(string separator) {
			return string.Join(separator, Array.ConvertAll(parameters, x => x.ToString()));
		}
		/// <summary>
		/// 適当文字列化
		/// </summary>
		public override string ToString() {
			return "{" + ParametersToString(", ") + "}";
		}
	}
}
