﻿using System;
using System.Data;
using System.Data.Common;
using System.Collections;

namespace Toolkit.Data {
	/// <summary>
	/// IDbTransactionのラッパー。
	/// </summary>
	/// <remarks>
	/// Disposeでは、Commitがtrueならコミット、falseならロールバックする。
	/// </remarks>
	public interface IDBTransaction : IDisposable {
		/// <summary>
		/// コミットするならtrueをセットする。デフォルトはfalse
		/// </summary>
		bool Commit { get; set; }

		/// <summary>
		/// 念のためIDbTransactionの取得
		/// </summary>
		IDbTransaction Transaction { get; }
	}

	/// <summary>
	/// IDbCommandのラッパー。
	/// </summary>
	public interface IDBCommand : IDisposable {
		/// <summary>
		/// パラメータの削除
		/// </summary>
		void ClearParameters();

		/// <summary>
		/// パラメータの追加
		/// </summary>
		void AddParameter(string name, object value);

		/// <summary>
		/// 実行
		/// </summary>
		/// <returns>影響を受けたレコード数</returns>
		int Execute();

		/// <summary>
		/// 実行
		/// </summary>
		IDataReader ExecuteReader();

		/// <summary>
		/// １つだけ取得。select count(*) from ... とか用
		/// </summary>
		object GetOne();

		/// <summary>
		/// １行だけ取得。select * from ... where HogeID = ? とか用
		/// </summary>
		/// <returns>値の配列</returns>
		object[] GetRowArray();

		/// <summary>
		/// １行だけ取得。select * from ... where HogeID = ? とか用
		/// </summary>
		/// <returns>名前(System.String)と値のHashtable</returns>
		Hashtable GetRowHash();

		/// <summary>
		/// SELECTなクエリの実行
		/// </summary>
		/// <returns>データテーブル</returns>
		DataTable GetAll();

		/// <summary>
		/// 念のためIDbCommandの取得
		/// </summary>
		IDbCommand Command { get; }
	}

	/// <summary>
	/// DB周りの簡単なラッパー。
	/// 直接使ってもいいものの、CreateCommand()やらCreateParameter()やら使うと微妙に面倒なので。。
	/// </summary>
	public class DBConnect : IDisposable {
		/// <summary>
		/// コネクション。
		/// </summary>
		IDbConnection con;

		/// <summary>
		/// とらんざくしょん。
		/// </summary>
		IDbTransaction tr = null;

		/// <summary>
		/// 初期化。
		/// </summary>
		public DBConnect(IDbConnection con) {
			this.con = con;
			this.con.Open();
		}

		/// <summary>
		/// 後始末。
		/// </summary>
		public void Dispose() {
			con.Close();
		}

		/// <summary>
		/// トランザクションの開始
		/// </summary>
		/// <example>
		/// using (DBConnect.Transaction tr = db.Begin()) {
		///     処理色々
		///     // 正常終了時
		///     tr.Commit = true;
		/// }
		/// </example>
		public IDBTransaction Begin() {
			return new DBTransaction(this);
		}

		/// <summary>
		/// トランザクションの開始
		/// </summary>
		/// <example>
		/// using (DBConnect.Transaction tr = db.Begin()) {
		///     処理色々
		///     // 正常終了時
		///     tr.Commit = true;
		/// }
		/// </example>
		public IDBTransaction Begin(IsolationLevel il) {
			return new DBTransaction(this, il);
		}

		/// <summary>
		/// SQLを実行
		/// </summary>
		/// <example>
		/// using (DBConnect.SqlCommand cmd = db.Query(sql)) {
		///     cmd.AddParameter("?Hoge", hoge);
		///     cmd.Execute();
		/// }
		/// </example>
		public IDBCommand Query(string sql) {
			return new DBCommand(this, sql);
		}

		/// <summary>
		/// 念のためIDbConnectionの取得
		/// </summary>
		public IDbConnection Connection {
			get { return con; }
		}

		#region IDBTransactionの実装

		/// <summary>
		/// トランザクションな処理
		/// </summary>
		class DBTransaction : IDBTransaction {
			bool commit = false;
			DBConnect db;
			public DBTransaction(DBConnect db) {
				this.db = db;
				this.db.tr = this.db.con.BeginTransaction();
			}

			public DBTransaction(DBConnect db, IsolationLevel il) {
				this.db = db;
				this.db.tr = this.db.con.BeginTransaction(il);
			}

			#region IDBTransaction メンバ

			public void Dispose() {
				if (commit) {
					db.tr.Commit();
				} else {
					db.tr.Rollback();
				}
				db.tr.Dispose();
				db.tr = null;
			}

			public bool Commit {
				get { return commit; } // 要らん気もするけど一応。
				set { commit = value; }
			}

			public IDbTransaction Transaction {
				get { return db.tr; }
			}

			#endregion
		}
		
		#endregion

		#region IDBCommandの実装
			
		/// <summary>
		/// IDbCommandのラッパー。
		/// </summary>
		class DBCommand : IDBCommand {
			IDbCommand cmd;
			public DBCommand(DBConnect db, string sql) {
				cmd = db.con.CreateCommand();
				if (db.tr != null) {
					cmd.Transaction = db.tr;
				}
				cmd.CommandText = sql;
				cmd.CommandType = CommandType.Text;
			}

			#region IDBCommand メンバ

			public void Dispose() {
				cmd.Dispose();
			}

			public void ClearParameters() {
				cmd.Parameters.Clear();
			}

			public void AddParameter(string name, object value) {
				IDbDataParameter param = cmd.CreateParameter();
				param.ParameterName = name;
				param.Value = value;
				cmd.Parameters.Add(param);
			}

			public int Execute() {
				return cmd.ExecuteNonQuery();
			}

			public IDataReader ExecuteReader() {
				return cmd.ExecuteReader();
			}

			public object GetOne() {
				using (IDataReader reader = cmd.ExecuteReader(CommandBehavior.SingleRow)) {
					if (reader.Read() && 0 < reader.FieldCount) {
						return reader.GetValue(0);
					}
					return null;
				}
			}

			public object[] GetRowArray() {
				using (IDataReader reader = cmd.ExecuteReader(CommandBehavior.SingleRow)) {
					if (reader.Read()) {
						object[] row = new object[reader.FieldCount];
						for (int i = 0; i < reader.FieldCount; i++) {
							row[i] = reader.GetValue(i);
						}
						return row;
					}
					return null;
				}
			}

			public Hashtable GetRowHash() {
				using (IDataReader reader = cmd.ExecuteReader(CommandBehavior.SingleRow)) {
					if (reader.Read()) {
						Hashtable row = new Hashtable();
						for (int i = 0; i < reader.FieldCount; i++) {
							row[reader.GetName(i)] = reader.GetValue(i);
						}
						return row;
					}
					return null;
				}
			}

			public DataTable GetAll() {
				using (IDataReader reader = cmd.ExecuteReader()) {
					DataTable table = new DataTable();
					// カラムの準備
					for (int i = 0; i < reader.FieldCount; i++) {
						table.Columns.Add(reader.GetName(i));
						// DuplicateNameExceptionはどーしたものか。。
					}
					// データをコピー
					while (reader.Read()) {
						object[] row = new object[reader.FieldCount];
						for (int i = 0; i < reader.FieldCount; i++) {
							row[i] = reader.GetValue(i);
						}
						table.Rows.Add(row);
					}
					return table;
				}
			}

			public IDbCommand Command {
				get { return cmd; }
			}

			#endregion
		}

		#endregion
	}
}
