﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace Toolkit.Windows {
	/// <summary>
	/// アイコン読み込んだぞイベントの引数。
	/// 失敗時は Icon == null
	/// </summary>
	public class IconLoadedEventArgs : EventArgs {
		public readonly System.Drawing.Icon Icon;
		public string FileName;
		public bool Small;
		public readonly object Arg;
		public IconLoadedEventArgs(System.Drawing.Icon icon, string fileName, bool small, object arg) {
			Icon = icon;
			FileName = fileName;
			Small = small;
			Arg = arg;
		}
	}

	/// <summary>
	/// アイコン読み込みスレッド。
	/// </summary>
	public class AsyncIconLoader {
		object lockObject = new object();
		volatile bool valid = false;
		Thread thread = null;
		Queue<Request> queue = new Queue<Request>();
		struct Request {
			public string FileName;
			public bool Small;
			public object Arg;
			public Request(string fileName, bool small, object arg) {
				FileName = fileName;
				Small = small;
				Arg = arg;
			}
		}

		/// <summary>
		/// アイコン読み込んだぞイベント
		/// </summary>
		public event EventHandler<IconLoadedEventArgs> IconLoaded;


		/// <summary>
		/// キューをクリアしてスレッド停止
		/// </summary>
		public void Clear() {
			Thread thread = null;
			lock (lockObject) {
				if (valid) {
					valid = false;
					Monitor.PulseAll(lockObject);
					thread = this.thread;
				}
			}
			if (thread != null) {
				thread.Join(1 * 1000);
				thread.Abort();
			}
		}

		/// <summary>
		/// 読み込み。
		/// </summary>
		public void Load(string fileName, bool small, object arg) {
			lock (lockObject) {
				if (!valid) {
					valid = true;
					thread = new Thread(new ThreadStart(OnThread));
					thread.SetApartmentState(ApartmentState.STA);
					thread.Start();
				}
				queue.Enqueue(new Request(fileName, small, arg));
				Monitor.PulseAll(lockObject);
			}
		}

		/// <summary>
		/// アイコン読み込みスレッド
		/// </summary>
		void OnThread() {
			while (valid) {
				// Dequeue
				Request r;
				lock (lockObject) {
					if (queue.Count <= 0) {
						Monitor.Wait(lockObject);
						continue;
					} else {
						r = queue.Dequeue();
					}
				}

				// アイコン読み込み
				try {
					using (System.Drawing.Icon icon = Icon.ExtractAssociatedIcon(
						 Toolkit.IO.Utility.PathNormalize(r.FileName), r.Small)) {

						if (!valid) {
							Toolkit.Windows.Icon.ReleaseIcon(icon);
							break;
						}

						CallEvent(r, icon);

						Toolkit.Windows.Icon.ReleaseIcon(icon);
					}
				} catch (System.IO.FileLoadException e) {
					System.Diagnostics.Debug.WriteLine(e.ToString());

					CallEvent(r, null);
				}
			}
		}

		private void CallEvent(Request r, System.Drawing.Icon icon) {
			EventHandler<IconLoadedEventArgs> IconLoaded = this.IconLoaded;
			if (IconLoaded != null) {
				IconLoaded(this, new IconLoadedEventArgs(
					icon, r.FileName, r.Small, r.Arg));
			}
		}
	}
}
